import { source } from 'sveltekit-sse';
import {
	decodeMessage,
	unmarshalMessage,
	type AMSError,
	type AMSStatus,
	type MarshalledMessage,
	type ShuntLog,
	type SlaveLog,
	type SlaveStatus
} from './messages';
import { writable, type Writable } from 'svelte/store';
import { SlaveLogData } from './slave-log';

export type ShuntData = {
	current: number;
	voltage1: number;
	voltage2: number;
};

export const error: Writable<AMSError | undefined> = writable(undefined);
export const amsStatus: Writable<AMSStatus | undefined> = writable(undefined);
export const slaveStatus: Writable<Record<number, SlaveStatus>> = writable({});
export const slaveLog: Writable<Record<number, SlaveLogData>> = writable({});
export const shunt: Writable<ShuntData> = writable({ current: 0, voltage1: 0, voltage2: 0 });

source('/data')
	.select('message')
	.subscribe((value) => {
		if (!value) {
			return;
		}
		const msg = decodeMessage(unmarshalMessage(JSON.parse(value) as MarshalledMessage));
		if (!msg) {
			return;
		}

		switch (msg.type) {
			case 'error':
				error.set(msg as AMSError);
				break;
			case 'status':
				amsStatus.set(msg as AMSStatus);
				break;
			case 'slaveStatus': {
				const status = msg as SlaveStatus;
				slaveStatus.update((r) => {
					r[status.slaveId] = status;
					return r;
				});
				break;
			}
			case 'slaveLog':
				handleSlaveLog(msg as SlaveLog);
				break;
			case 'shuntCurrentLog':
				shunt.update((s) => {
					s.current = (msg as ShuntLog).value;
					return s;
				});
				break;
			case 'shuntVoltage1Log':
				shunt.update((s) => Object.assign(s, { voltage1: (msg as ShuntLog).value }));
				break;
			case 'shuntVoltage2Log':
				shunt.update((s) => Object.assign(s, { voltage2: (msg as ShuntLog).value }));
				break;
		}
	});

function handleSlaveLog(msg: SlaveLog) {
	slaveLog.update((r) => {
		if (!(msg.slaveId in r)) {
			r[msg.slaveId] = new SlaveLogData();
		}
		const slave = r[msg.slaveId];
		slave.handleMsg(msg);
		return r;
	});
}
