export interface MarshalledMessage {
	id: number;
	ext: boolean;
	canfd: boolean;
	data: number[];
}

export type RawMessage = {
	ext?: boolean;
	canfd?: boolean;
	id: number;
	data: Buffer;
};

export function marshalMessage(msg: RawMessage): MarshalledMessage {
	return {
		id: msg.id,
		ext: msg.ext ?? false,
		canfd: msg.canfd ?? false,
		data: Array.from(msg.data)
	};
}

export function unmarshalMessage(msg: MarshalledMessage): RawMessage {
	return {
		id: msg.id,
		ext: msg.ext,
		canfd: msg.canfd,
		data: Buffer.from(msg.data)
	};
}

export interface AMSMessage {
	type:
		| 'error'
		| 'status'
		| 'slaveStatus'
		| 'slaveLog'
		| 'shuntCurrentLog'
		| 'shuntVoltage1Log'
		| 'shuntVoltage2Log';
}

export interface AMSError extends AMSMessage {
	type: 'error';

	kind: number;
	arg: number;
}

export enum AMSState {
	TS_INACTIVE = 0,
	TS_ACTIVE = 1,
	TS_PRECHARGE = 2,
	TS_DISCHARGE = 3,
	TS_ERROR = 4
}

export interface AMSStatus extends AMSMessage {
	type: 'status';

	state: AMSState;
	sdcClosed: boolean;
	soc: number;
	minCellVolt: number;
	maxCellTemp: number;
	imdOK: boolean;
	// TODO: IMD state & R_iso
}

export interface SlaveStatus extends AMSMessage {
	type: 'slaveStatus';

	slaveId: number;
	error: boolean;
	minCellVolt: number;
	maxCellVolt: number;
	maxTemp: number;
	soc: number;
}

export interface SlaveLog extends AMSMessage {
	type: 'slaveLog';

	slaveId: number;

	logType: 'voltage' | 'lastCell' | 'temperature';
}

export interface SlaveLogVoltage extends SlaveLog {
	logType: 'voltage';

	startIndex: number;
	voltages: [number, number, number, number];
}

export interface SlaveLogLastCell extends SlaveLog {
	logType: 'lastCell';

	voltage: number;
	failed_temp_sensors: number;
}

export interface SlaveLogTemperature extends SlaveLog {
	logType: 'temperature';

	startIndex: number;
	temperatures: [number, number, number, number, number, number, number, number];
}

export interface ShuntLog extends AMSMessage {
	type: 'shuntCurrentLog' | 'shuntVoltage1Log' | 'shuntVoltage2Log';

	value: number;
}

import { N_CELLS_PER_SLAVE, N_TEMP_SENSORS_PER_SLAVE } from '$lib/defs';

export const CAN_ID_AMS_ERROR = 0x00c;
export const CAN_ID_AMS_STATUS = 0x00a;
export const CAN_ID_AMS_SLAVE_STATUS_BASE = 0x080;
export const CAN_ID_AMS_SLAVE_STATUS_MASK = 0xff0;
export const CAN_ID_AMS_SLAVE_LOG_BASE = 0x600;
export const CAN_ID_AMS_SLAVE_LOG_MASK = 0xf00;
export const CAN_ID_SHUNT_CURRENT = 0x521;
export const CAN_ID_SHUNT_VOLTAGE_1 = 0x522;
export const CAN_ID_SHUNT_VOLTAGE_2 = 0x523;

function decodeError(msg: RawMessage): AMSError | null {
	if (msg.data.length != 2) {
		console.warn('invalid error frame', msg);
		return null;
	}

	console.log('error', msg);
	const data = msg.data;
	return {
		type: 'error',
		kind: data[0],
		arg: data[1]
	};
}

function decodeStatus(msg: RawMessage): AMSStatus | null {
	if (msg.data.length != 8) {
		console.warn('invalid status frame', msg);
		return null;
	}

	const data = msg.data;
	return {
		type: 'status',
		state: data[0] & 0x7f,
		sdcClosed: !!(data[0] & 0x80),
		soc: data[1],
		minCellVolt: data.readUInt16BE(2) * 1e-3,
		maxCellTemp: data.readUInt16BE(4) * 0.0625,
		imdOK: !!(data[6] & 0x80)
	};
}

function decodeSlaveStatus(msg: RawMessage): SlaveStatus | null {
	if (msg.data.length != 8) {
		console.warn('invalid slave status frame', msg);
		return null;
	}

	const data = msg.data;
	return {
		type: 'slaveStatus',

		slaveId: data[0] & 0x7f,
		error: !!(data[0] & 0x80),
		soc: data[1],
		minCellVolt: data.readUInt16BE(2) * 1e-3,
		maxCellVolt: data.readUInt16BE(4) * 1e-3,
		maxTemp: data.readUInt16BE(6) * 0.0625
	};
}

function decodeSlaveLog(msg: RawMessage): SlaveLog | null {
	if (msg.data.length != 8) {
		console.warn('invalid slave log frame', msg);
		return null;
	}

	const slaveId = (msg.id & 0xf0) >> 4;
	const logIndex = msg.id & 0x00f;
	const data = msg.data;
	if (logIndex < N_CELLS_PER_SLAVE / 4) {
		const startIndex = logIndex * 4;
		const voltages = [];
		for (let i = 0; i < 4; i++) {
			voltages.push(data.readUInt16BE(2 * i) * 1e-3);
		}
		const msg: SlaveLogVoltage = {
			type: 'slaveLog',
			slaveId,

			logType: 'voltage',
			startIndex,
			voltages: voltages as [number, number, number, number]
		};
		return msg;
	} else if (logIndex == N_CELLS_PER_SLAVE / 4) {
		const msg: SlaveLogLastCell = {
			type: 'slaveLog',
			slaveId,

			logType: 'lastCell',
			voltage: data.readUInt16BE(0) * 1e-3,
			failed_temp_sensors: data.readInt32BE(2)
		};
		return msg;
	} else if (logIndex < N_CELLS_PER_SLAVE / 4 + 1 + N_TEMP_SENSORS_PER_SLAVE / 8) {
		const temperatures = [];
		for (let i = 0; i < 8; i++) {
			temperatures.push(data.readInt8(i) * 1);
		}
		const msg: SlaveLogTemperature = {
			type: 'slaveLog',
			slaveId,

			logType: 'temperature',
			startIndex: (logIndex - N_CELLS_PER_SLAVE / 4 - 1) * 8,
			temperatures: temperatures as [number, number, number, number, number, number, number, number]
		};
		return msg;
	} else {
		console.warn('Unknown slave log index', msg.id);
		return null;
	}
}

function decodeShunt(msg: RawMessage, type: ShuntLog['type']): ShuntLog | null {
	if (msg.data.length != 6) {
		console.warn('invalid shunt log frame', msg);
		return null;
	}

	return {
		type: type,
		value: msg.data.readInt32BE(2) * 1e-3
	};
}

export function decodeMessage(msg: RawMessage): AMSMessage | null {
	if (msg.ext || msg.canfd) {
		console.warn('invalid message', msg);
		return null;
	}

	let message: AMSMessage | null = null;

	if ((msg.id & CAN_ID_AMS_SLAVE_STATUS_MASK) == CAN_ID_AMS_SLAVE_STATUS_BASE) {
		message = decodeSlaveStatus(msg);
	} else if ((msg.id & CAN_ID_AMS_SLAVE_LOG_MASK) == CAN_ID_AMS_SLAVE_LOG_BASE) {
		message = decodeSlaveLog(msg);
	} else {
		switch (msg.id) {
			case CAN_ID_AMS_ERROR:
				message = decodeError(msg);
				break;
			case CAN_ID_AMS_STATUS:
				message = decodeStatus(msg);
				break;
			case CAN_ID_SHUNT_CURRENT:
				message = decodeShunt(msg, 'shuntCurrentLog');
				break;
			case CAN_ID_SHUNT_VOLTAGE_1:
				message = decodeShunt(msg, 'shuntVoltage1Log');
				break;
			case CAN_ID_SHUNT_VOLTAGE_2:
				message = decodeShunt(msg, 'shuntVoltage2Log');
				break;
			default:
				console.warn(`unknown message id ${msg.id}`, msg);
		}
	}

	return message;
}
